<?php

    namespace pachno\modules\livelink_github;

    use Github\ResultPager;
    use pachno\core\entities\Branch;
    use pachno\core\entities\BranchCommit;
    use pachno\core\entities\Comment;
    use pachno\core\entities\Commit;
    use pachno\core\entities\CommitFile;
    use pachno\core\entities\Issue;
    use pachno\core\entities\Issuetype;
    use pachno\core\entities\LivelinkImport;
    use pachno\core\entities\Milestone;
    use pachno\core\entities\Project;
    use pachno\core\entities\Status;
    use pachno\core\entities\tables\BranchCommits;
    use pachno\core\entities\tables\Branches;
    use pachno\core\entities\tables\Commits;
    use pachno\core\entities\tables\Modules;
    use pachno\core\entities\tables\Users;
    use pachno\core\entities\User;
    use pachno\core\framework,
        Github\Client as GithubClient,
        pachno\core\modules\livelink\ConnectorProvider,
        pachno\core\modules\livelink\Livelink;
    use pachno\core\framework\cli\Command;
    use pachno\core\framework\exceptions\SettingsException;
    use pachno\core\framework\Request;

    /**
     * Pachno Github integration module
     *
     * @package livelink
     * @subpackage core
     *
     * @Table(name="\pachno\core\entities\tables\Modules")
     */
    class Livelink_github extends \pachno\core\entities\Module implements ConnectorProvider
    {

        public const VERSION = '0.1';

        protected $_name = 'livelink_github';
        protected $_longname = 'GitHub integration';
        protected $_description = 'The GitHub integration module provides live integration and updates between Pachno and projects in GitHub';
        protected $_module_config_title = 'GitHub integration';
        protected $_module_config_description = 'Set up the GitHub integration module from this section';
        protected $_has_account_settings = false;
        protected $_account_settings_name = '';
        protected $_account_settings_logo = '';
        protected $_has_composer_dependencies = true;
        protected $_composer_name = 'pachno/livelink_github';

        /**
         * @var Connector
         */
        protected $_connector;

        /**
         * Return an instance of this module
         *
         * @return Livelink_github
         */
        public const SETTINGS_REPOSITORY_URL_PROJECT = 'repository_url_project_';

        public static function getModule()
        {
            return framework\Context::getModule('livelink_github');
        }

        protected function _initialize()
        {
            Livelink::getModule()->addConnector('github', $this);
        }

        protected function _addListeners()
        {
        }

        protected function _install($scope)
        {
        }

        protected function _loadFixtures($scope)
        {
        }

        protected function _uninstall()
        {
        }

        /**
         * @return \pachno\modules\livelink_github\Connector
         */
        public function getConnector()
        {
            if (!$this->_connector instanceof Connector) {
                $this->_connector = new Connector();
            }

            return $this->_connector;
        }

        public function getInputOptionsForProjectEdit(framework\Request $request)
        {
            $inputs = [
                'repository_url' => $request['repository_url'],
                'live_link' => $request['live_link'],
                'connector' => 'github'
            ];

            return $inputs;
        }

        public function getNormalizedRepositoryDetails($repository_url)
        {
            $url = str_replace(['git@github.com:', 'https://github.com/'], ['', ''], $repository_url);
            $pieces = parse_url($url);
            $pieces = explode('/', $pieces['path']);
            $repository_user = '';
            $repository_name = '';
            if (count($pieces) == 2) {
                list($repository_user, $repository_name) = $pieces;
                if (substr($repository_name, -4) == '.git') {
                    $repository_name = substr($repository_name, 0, -4);
                }
            }

            return [$repository_user, $repository_name];
        }

        public function getRepositoryDisplayNameForProject(Project $project)
        {
            $repository_url = $this->getProjectRepositoryUrl($project);
            list ($repository_user, $repository_name) = $this->getNormalizedRepositoryDetails($repository_url);
            return $repository_user . '/' . $repository_name;
        }

        public function getImportDisplayNameForProjectEdit(framework\Request $request)
        {
            list ($repository_user, $repository_name) = $this->getNormalizedRepositoryDetails($request['repository_url']);
            return $repository_user . '/' . $repository_name;
        }

        public function getImportProjectNameForProjectEdit(framework\Request $request)
        {
            list ($repository_user, $repository_name) = $this->getNormalizedRepositoryDetails($request['repository_url']);
            return ucfirst($repository_name);
        }

        public function getProjectRepositoryUrl(Project $project)
        {
            return $this->getSetting(self::SETTINGS_REPOSITORY_URL_PROJECT . $project->getID());
        }

        protected function getUserId($username)
        {
            static $user_cache;

            if (!is_array($user_cache)) {
                $user_cache = [];
            }

            if (!isset($user_cache[$username])) {
                $user = Users::getTable()->getByUsername($username);
                if ($user instanceof User) {
                    $user_cache[$username] = $user->getID();
                } else {
                    $user_details = $this->getConnector()->getGithubClient()->api('user')->show($username);

                    if ($user_details['email']) {
                        $user = Users::getTable()->getByEmail($user_details['email']);
                    }

                    if (!$user instanceof User) {
                        $user = new User();
                        $user->setUsername($username);
                        $user->setRealname($user_details['name']);
                        $user->setBuddyname($username);
                        if ($user_details['email']) {
                            $user->setEmail($user_details['email']);
                        }
                        $user->save();
                    }

                    $user_cache[$username] = $user->getID();
                }
            }

            return $user_cache[$username] ?? null;
        }

        /**
         * @param $labels
         * @param Issuetype[] $issuetypes
         * @return Issuetype
         */
        protected function getIssueTypeFromLabels($labels, $issuetypes)
        {
            foreach ($labels as $label) {
                foreach ($issuetypes as $item) {
                    $key = $item->getKey();
                    $normalized = str_replace(['-', '_', ' ', '/', "'"], ['', '', '', '', ''], mb_strtolower($label['name']));

                    if ($key == $normalized || stripos($key, $normalized) !== false || stripos($normalized, $key) !== false) {
                        return $item;
                    }
                }
            }

            return array_values($issuetypes)[0];
        }

        /**
         * @param $labels
         * @param Status[] $statuses
         * @return Status
         */
        protected function getStatusFromLabels($labels, $statuses)
        {
            foreach ($labels as $label) {
                foreach ($statuses as $key => $item) {
                    if ($key == str_replace(['-', '_', ' ', '/', "'"], ['', '', '', '', ''], mb_strtolower($label['name']))) {
                        return $item;
                    }
                }
            }

            return array_values($statuses)[0];
        }

        /**
         * @param Project $project
         * @param $data
         * @param string[] $labels
         * @param Milestone[] $milestones
         * @param Status[] $statuses
         * @param Issuetype[] $issuetypes
         *
         * @return Issue
         * @throws \Exception
         */
        protected function createIssueFromData(Project $project, $data, $labels, $milestones, $statuses, $issuetypes)
        {
            $issue = new Issue();
            $issue->setProject($project);
            $issue->setTitle($data['title']);
            $issue->setDescription($data['body']);
            $issue->setIssueNo($data['number']);
            $issue->setDescriptionSyntax(framework\Settings::SYNTAX_MD);
            $issue->setPostedBy($this->getUserId($data['user']['login']));
            $created_at = new \DateTime($data['created_at']);
            $issue->setPosted($created_at->getTimestamp());
            $issue->setIssuetype($this->getIssueTypeFromLabels($labels, $issuetypes)->getID());
            if (isset($data['assignee']) && isset($data['assignee']['login'])) {
                $issue->setAssignee(Users::getTable()->getByUserID($this->getUserId($data['assignee']['login'])));
            }
            $status = $this->getStatusFromLabels($labels, $statuses);
            if ($status instanceof Status) {
                $issue->setStatus($status);
            }

            if ($data['milestone'] && isset($data['milestone']['number']) && isset($milestones[$data['milestone']['number']])) {
                $issue->setMilestone($milestones[$data['milestone']['number']]);
            }

            $issue->save();

            return $issue;
        }

        protected function importBranch(GithubClient $client, Project $project, User $user, $repository_user, $repository_name, $branch_data)
        {
            $branch = new Branch();
            $branch->setName($branch_data['name']);
            $branch->setProject($project);
            $branch->save();

            $this->importCommits($client, $project, $user, $branch, $repository_user, $repository_name);
        }

        protected function importBranches(GithubClient $client, Project $project, User $user, $repository_user, $repository_name)
        {
            if (framework\Context::isCLI()) {
                Command::cli_echo("Importing branches\n");
            }

            $master = $client->api('repo')->branches($repository_user, $repository_name, 'master');
            $this->importBranch($client, $project, $user, $repository_user, $repository_name, $master);
            framework\Context::switchUserContext($user);

            $branches = $client->api('repo')->branches($repository_user, $repository_name);
            foreach ($branches as $branch_data) {
                if ($branch_data['name'] == 'master') {
                    continue;
                }
                $this->importBranch($client, $project, $user, $repository_user, $repository_name, $branch_data);

                framework\Context::switchUserContext($user);
            }
        }

        public function importSingleCommit(Project $project, Commit $commit)
        {
            $repository_url = $this->getProjectRepositoryUrl($project);
            if (!$repository_url) {
                throw new \Exception('This project does not have a repository url stored');
            }
            list($repository_user, $repository_name) = $this->getNormalizedRepositoryDetails($repository_url);

            $client = $this->getConnector()->getGithubClient();

            $commit_data = $client->api('repo')->commits()->show($repository_user, $repository_name, $commit->getRevision());
            $parent_commit = $commit_data['parents'][0]['sha'] ?? '';
            $branch = new Branch();

            $commit = $this->importCommit($project, $branch, $commit_data, $parent_commit, false, true);
            $commit->setIsImported();
            $commit->save();
        }

        protected function importCommits(GithubClient $client, Project $project, User $user, Branch $branch, $repository_user, $repository_name)
        {
            $commitsapi = $client->api('repo')->commits();
            $paginator = new ResultPager($client);

            if (framework\Context::isCLI()) {
                Command::cli_echo("Importing commits for branch ");
                Command::cli_echo($branch->getName() . "\n", 'white', 'bold');
            }
            $commits = $paginator->fetch($commitsapi, 'all', [$repository_user, $repository_name, ['sha' => $branch->getName()]]);
            $first = true;
            while (count($commits)) {
                if (framework\Context::isCLI()) {
                    Command::cli_echo('.');
                }
                foreach ($commits as $commit) {
                    $parent_commit = $commit['parents'][0]['sha'] ?? '';

                    $commit = $this->importCommit($project, $branch, $commit, $parent_commit, $first);
                    framework\Context::switchUserContext($user);
                    $first = false;
                }

                if ($paginator->hasNext()) {
                    $commits = $paginator->fetchNext();
                } else {
                    $commits = [];
                }
            }
            if (framework\Context::isCLI()) {
                Command::cli_echo(" done!\n");
            }
        }

        protected function createCommentsForIssue(Issue $issue, $data, $repository_user, $repository_name)
        {
            $client = $this->getConnector()->getGithubClient();
            $commentsapi = $client->api('issue')->comments();
            $paginator = new ResultPager($client);

            if (framework\Context::isCLI()) {
                Command::cli_echo("Importing comment(s)");
            }
            $comments = $paginator->fetchAll($commentsapi, 'all', [$repository_user, $repository_name, $data['number']]);
            if (framework\Context::isCLI()) {
                Command::cli_echo(" (" . count($comments) . ")\n");
            }

            foreach ($comments as $comment_data) {
                try {
                    $comment = new Comment();
                    $comment->setTargetType(Comment::TYPE_ISSUE);
                    $comment->setTargetID($issue->getID());
                    $comment->setContent($comment_data['body']);
                    $comment->setSyntax(framework\Settings::SYNTAX_MD);
                    $comment->setPostedBy($this->getUserId($comment_data['user']['login']));
                    $created_at = new \DateTime($comment_data['created_at']);
                    $comment->setPosted($created_at->getTimestamp());
                    $comment->save();
                } catch (\b2db\Exception $e) {
                    if (framework\Context::isCLI()) {
                        Command::cli_echo("Could not import comment " . $comment_data['id'] . " (" . $e->getMessage() . ")\n", 'red');
                    }
                }
            }
        }

        /**
         * @param Project $project
         * @param $repository_user
         * @param $repository_name
         * @return Milestone[]
         * @throws \Exception
         */
        protected function importMilestones(GithubClient $client, Project $project, $repository_user, $repository_name)
        {
            $milestones = [];
            $milestonesapi = $client->api('issue')->milestones();
            $paginator = new ResultPager($client);
            $milestones_data = $paginator->fetchAll($milestonesapi, 'all', [$repository_user, $repository_name]);

            foreach ($milestones_data as $milestone_data) {
                $milestone = new Milestone();
                $milestone->setProject($project);
                $milestone->setName($milestone_data['title']);
                $milestone->setDescription($milestone_data['description']);
                $milestone->save();

                $milestones[$milestone_data['number']] = $milestone;
            }

            return $milestones;
        }

        protected function importIssues(GithubClient $client, Project $project, $repository_user, $repository_name)
        {
            $issuesapi = $client->api('issue');
            $paginator = new ResultPager($client);
            $issues = $paginator->fetchAll($issuesapi, 'all', [$repository_user, $repository_name, ['state' => 'open']]);
            $milestones = $this->importMilestones($client, $project, $repository_user, $repository_name);
            $statuses = $project->getAvailableStatuses();
            $issuetypes = $project->getIssuetypeScheme()->getIssuetypes();
            $current = 1;
            $current_data = null;

            foreach ($issues as $data) {
                try {
                    if (framework\Context::isCLI()) {
                        $current_data = $data;
                        Command::cli_echo("Importing issue {$current} of ".count($issues)."\n");
                    }
                    $labels = $client->api('issue')->labels()->all($repository_user, $repository_name, $data['number']);
                    $issue = $this->createIssueFromData($project, $data, $labels, $milestones, $statuses, $issuetypes);

                    $this->createCommentsForIssue($issue, $data, $repository_user, $repository_name);

                    $updated_at = new \DateTime($data['updated_at']);
                    $issue->touch($updated_at->getTimestamp());
                    $current++;
                } catch (\Exception $e) {
                    if (framework\Context::isCLI()) {
                        Command::cli_echo("Data error:\n");
                        Command::cli_echo(print_r($data, true));
                        Command::cli_echo("---\n");
                    }
                }
            }
        }

        public function importProject(Project $project, User $user)
        {
            $repository_url = $this->getProjectRepositoryUrl($project);
            if (!$repository_url) {
                throw new \Exception('This project does not have a repository url stored');
            }
            list($repository_user, $repository_name) = $this->getNormalizedRepositoryDetails($repository_url);

            $client = $this->getConnector()->getGithubClient();
            $this->importIssues($client, $project, $repository_user, $repository_name);
            $this->importBranches($client, $project, $user, $repository_user, $repository_name);
            $this->updateCommits($project);
        }

        protected function updateCommits(Project $project)
        {
            if (framework\Context::isCLI()) {
                Command::cli_echo("Fixing commit relations ...");
            }
            $commits = Commits::getTable()->getUnprocessedCommitsByProject($project);

            foreach ($commits as $commit) {
                $previous_commit = Commits::getTable()->getCommitByHash($commit->getPreviousRevision(), $project);
                if ($previous_commit instanceof Commit) {
                    $commit->setPreviousCommit($previous_commit);
                    $commit->save();
                }
            }
            if (framework\Context::isCLI()) {
                Command::cli_echo(" done!\n");
            }
        }

        public function removeProjectConnectorSettings(Project $project, $secret)
        {
            if (Livelink::getModule()->isLiveLinkEnabledForProject($project)) {
                $repository_url = $this->getProjectRepositoryUrl($project);
                list ($repository_user, $repository_name) = $this->getNormalizedRepositoryDetails($repository_url);
                $client = $this->getConnector()->getGithubClient();
                $url = framework\Context::getRouting()->generate('livelink_webhook', ['project_id' => $project->getID(), 'secret' => $secret], false);
                $hooks = $client->api('repo')->hooks()->all($repository_user, $repository_name);

                foreach ($hooks as $hook) {
                    if (isset($hook['config']) && isset($hook['config']['url']) && $hook['config']['url'] == $url) {
                        $client->api('repo')->hooks()->remove($repository_user, $repository_name, $hook['id']);
                    }
                }

                Livelink::getModule()->setLiveLinkEnabledForProject($project, false);
            }
            $this->deleteSetting(self::SETTINGS_REPOSITORY_URL_PROJECT . $project->getID());
        }

        public function saveProjectConnectorSettings(framework\Request $request, Project $project, $secret)
        {
            $livelink_mode = $request['live_link'];

            $this->saveSetting(self::SETTINGS_REPOSITORY_URL_PROJECT . $project->getID(), $request['repository_url']);

            if (in_array($livelink_mode, ['simple', 'full'])) {
                list ($repository_user, $repository_name) = $this->getNormalizedRepositoryDetails($request['repository_url']);
                $client = $this->getConnector()->getGithubClient();
                $hook = $client->api('repo')->hooks()->create($repository_user, $repository_name, [
                    'name' => 'web',
                    'config' => [
                        'url' => framework\Context::getRouting()->generate('livelink_webhook', ['project_id' => $project->getID(), 'secret' => $secret], false),
                        'content_type' => 'json'
                    ],
                    'events' => ['push', 'delete']
                ]);
                Livelink::getModule()->setLiveLinkEnabledForProject($project, true);
            }

            if (in_array($livelink_mode, ['import', 'full'])) {
                $import = new LivelinkImport();
                $import->setProject($project);
                $import->setUser(framework\Context::getUser());
                $import->save();
            }
        }

        public function postConnectorSettings(framework\Request $request)
        {
            $token = trim($request['github_token']);
            $client = new GithubClient();
            $client->authenticate($token, null, GithubClient::AUTH_HTTP_TOKEN);
            try {
                $client->currentUser()->emails()->all();
                framework\Settings::saveUserSetting(framework\Context::getUser()->getID(), 'github_token', $token, $this->getName());

                return ['connected' => true];
            } catch (\Exception $e) {
                if ($e->getCode() == 401) {
                    throw new \Exception('Could not authenticate with GitHub using the provided token');
                } else {
                    throw $e;
                }
            }
        }

        public function removeConnectorSettings(framework\Request $request)
        {
            try {
                framework\Settings::deleteUserSetting(framework\Context::getUser()->getID(), 'github_token', $this->getName());

                return ['removed' => true];
            } catch (\Exception $e) {
                if ($e->getCode() == 401) {
                    throw new \Exception('Could not authenticate with GitHub using the provided token');
                } else {
                    throw $e;
                }
            }
        }

        public function webhook(Request $request, Project $project)
        {
            $event_type = $_SERVER['HTTP_X_GITHUB_EVENT'];

            switch ($event_type) {
                case 'push':
                    return $this->handlePushWebhook($request, $project);

                case 'delete':
                    return $this->handleDeleteWebhook($request, $project);

            }
        }

        /**
         * Import a single commit
         *
         * @param Project $project
         * @param Branch $branch
         * @param $commit
         * @param null $previous_commit
         * @param bool $update_branch
         *
         * @return Commit
         */
        protected function importCommit(Project $project, Branch $branch, $commit, $previous_commit, $update_branch = true, $force = false)
        {
            $author = (is_array($commit)) ? $commit['commit']['author']['email'] : $commit->author->email;
            $current_commit = (is_array($commit)) ? $commit['sha'] : $commit->id;
            $message = (is_array($commit)) ? $commit['commit']['message'] : $commit->message;
            $date = (is_array($commit)) ? $commit['commit']['author']['date'] : $commit->author->date;
            $date = new \DateTime($date);
            $isProjectCommitProcessed = Commits::getTable()->isProjectCommitProcessed($current_commit, $project->getID());

            if ($branch->getID() && !BranchCommits::getTable()->hasBranchCommitSha($branch, $current_commit)) {
                if ($isProjectCommitProcessed) {
                    $branchCommit = new BranchCommit();
                    $commit_object = Commits::getTable()->getCommitByHash($current_commit, $project);
                    $branchCommit->setCommit($commit_object);
                    $branchCommit->setBranch($branch);
                    $branchCommit->save();
                }
            }

            if (!$force && $isProjectCommitProcessed) {
                $commit_object = Commits::getTable()->getCommitByHash($current_commit, $project);
                return $commit_object;
            }

            $changes = [];

            // Build arrays of affected files
            if (is_array($commit) && array_key_exists('files', $commit)) {
                foreach ($commit['files'] as $file) {
                    switch ($file['status']) {
                        case 'modified':
                            $changes[] = [
                                'action' => CommitFile::ACTION_MODIFIED,
                                'filename' => $file['filename'],
                                'diff' => $file['patch'] ?? ''
                            ];
                            break;
                        case 'added':
                            $changes[] = [
                                'action' => CommitFile::ACTION_ADDED,
                                'filename' => $file['filename'],
                                'diff' => $file['patch'] ?? ''
                            ];
                            break;
                        case 'renamed':
                            $changes[] = [
                                'action' => CommitFile::ACTION_RENAMED,
                                'previous_filename' => $file['previous_filename'],
                                'filename' => $file['filename'],
                                'diff' => $file['patch'] ?? ''
                            ];
                            break;
                        case 'removed':
                        case 'deleted':
                            $changes[] = [
                                'action' => CommitFile::ACTION_DELETED,
                                'filename' => $file['filename'],
                                'diff' => $file['patch'] ?? ''
                            ];
                            break;
                        default:
                            break;
                    }
                }
            } else {
                if (is_array($commit)) {
                    $added = (array_key_exists('added', $commit)) ? $commit['added'] : [];
                    $modified = (array_key_exists('modified', $commit)) ? $commit['modified'] : [];
                    $removed = (array_key_exists('removed', $commit)) ? $commit['removed'] : [];
                } else {
                    $added = (property_exists($commit, 'added')) ? $commit->added : [];
                    $modified = (property_exists($commit, 'modified')) ? $commit->modified : [];
                    $removed = (property_exists($commit, 'removed')) ? $commit->removed : [];
                }

                foreach ($added as $added_file) {
                    $changes[] = [
                        'action' => CommitFile::ACTION_ADDED,
                        'filename' => $added_file
                    ];
                }
                foreach ($modified as $modified_file) {
                    $changes[] = [
                        'action' => CommitFile::ACTION_MODIFIED,
                        'filename' => $modified_file
                    ];
                }
                foreach ($removed as $removed_file) {
                    $changes[] = [
                        'action' => CommitFile::ACTION_DELETED,
                        'filename' => $removed_file
                    ];
                }
            }

            // Add commit
            return Livelink::getModule()->processCommit($project, $branch, $message, $author, $date, $previous_commit, $current_commit, $changes, $update_branch, [], true);
        }

        protected function handleDeleteWebhook(Request $request, Project $project)
        {
            $payload = ($request->hasParameter('payload')) ? $request->getParameter('payload') : $request->getInput();
            $entry = json_decode($payload);

            if ($entry->ref_type == 'branch') {
                $branch_name = Branch::getBranchNameFromRef($entry->ref);
                $branch = Branches::getTable()->getByBranchNameAndProject($project, $branch_name);

                if ($branch instanceof Branch) {
                    $branch->setIsDeleted();
                    $branch->save();
                }
            }
        }

        protected function handlePushWebhook(Request $request, Project $project)
        {
            $payload = ($request->hasParameter('payload')) ? $request->getParameter('payload') : $request->getInput();
            $entries = json_decode($payload);
            if ($entries == null)
            {
                throw new \Exception('The webhook payload could not be decoded');
            }

            $previous_commit = $entries->before;

            // Branch is stored in the ref
            $branch_name = Branch::getBranchNameFromRef($entries->ref);
            $branch = Branches::getTable()->getOrCreateByBranchNameAndProject($branch_name, $project);

            // Parse each commit individually
            foreach ($entries->commits as $commit_data)
            {
                $commit = $this->importCommit($project, $branch, $commit_data, $previous_commit);
                $commit = $this->importSingleCommit($project, $commit);
                $commit->setIsImported();
                $commit->save();
                $previous_commit = $commit_data->id;
            }
        }

        public function getCommitUrl(Commit $commit)
        {
            $url = 'https://github.com/' .
                $this->getRepositoryDisplayNameForProject($commit->getProject()) .
                '/commit/' . $commit->getRevision();

            return $url;
        }

    }
